'''
This file contains pre-processing, post-processing, and analytical tools for simulations.

Public Funcions:
- check_convergence:    Check whether a simulation result converges. i.e. whether U, V's fluctuation are very small.
- combine_sim:          Combine two simulation objects and return a new one (the first two unchanged).
                        Intended usage: say you have sim1, sim2 with same parameters except for sim_time, say 10 and 20.
                                        Then combine_sim takes a weighted average (with ratio 1:2) of results and return a new sim3.
                                        So that you now have sim3 with 30 sim_time.

Private Functions:
- rounds_expected:      Roughly calculates how many rounds are expected in a single simulation (which reflects runtime).
                        NOTE: Not well-developed. Not recommending to use.
- scale_maxtime:        Given two simulation objects, scale first one's maxtime towards the second, so that the two have the same expected rounds.
                        Intended to possibly decrease maxtime and save runtime.
                        NOTE: Not well-developed. Not recommending to use.

'''

from . import model as model
from . import figures as figures
from .tools import figure_tools as figure_t

import numpy as np
import math




def rounds_expected(sim):
    '''
    NOTE: Not well-developed. Not recommending to use.

    Predict how many rounds will run in single_test. i.e., how many for loops from time = 0 to sim.maxtime.
    Calculated based on expected_UV. 
    '''

    N = sim.N
    M = sim.M
    U_expected, V_expected = figures.UV_expected_val(sim)

    rates = []
    patch0 = None  # simulate patch i, j
    patch0_nb = []  # simulate neighbors of patch i, j
    
    # loop through N, M, create a sample patch to calculate rates, store them
    for i in range(N):
        for j in range(M):
            patch0 = model.patch(U_expected[i][j], V_expected[i][j], sim.X[i][j], sim.P[i][j])

            nb_indices = None
            if sim.boundary:
                nb_indices = model.find_nb_zero_flux(N, M, i, j)
            else:
                nb_indices = model.find_nb_periodical(N, M, i, j)
            
            for k in range(4):
                if nb_indices[k] != None:
                    i_nb = nb_indices[k][0]
                    j_nb = nb_indices[k][1]
                    patch0_nb_k = model.patch(U_expected[i_nb][j_nb], V_expected[i_nb][j_nb], sim.X[i_nb][j_nb], sim.P[i_nb][j_nb])
                    patch0_nb_k.update_pi_k()
                    patch0_nb.append(patch0_nb_k)

                else:
                    patch0_nb.append(None)

            patch0.nb = patch0_nb
            patch0.update_pi_k()
            patch0.update_mig()

            rates += patch0.pi_death_rates
            rates += patch0.mig_rates
    
    delta_t_expected = (1 / sum(rates)) * math.log(1 / 0.5)
    r_expected = round(sim.maxtime / delta_t_expected)

    return r_expected




def scale_maxtime(sim1, sim2, scale_interval = True):
    '''
    NOTE: Not well-developed. Not recommending to use.

    Scale sim1's maxtime towards sim2's, so they will run similar number of rounds in single_test, and hence similar runtime.
    Intended to reduce the effect of changing params on runtime. 
    
    Input:
    - scale_interval decides whether to scale sim1's interval as well, so that the same number of data will be stored.
    '''

    r_expected1 = rounds_expected(sim1)
    r_expected2 = rounds_expected(sim2)
    ratio = r_expected2 / r_expected1

    new_maxtime = sim1.maxtime * ratio
    old_max_record = sim1.maxtime / sim1.interval

    if scale_interval:
        sim1.interval = new_maxtime / old_max_record

    sim1.change_maxtime(new_maxtime)




def check_convergence(sim, interval = 20, start = 0.8, fluc = 0.07):
    '''
    Check whether a simulation converges or not.
    Based on whether the fluctuation of U, V, pi all < 'fluc' in the later 'tail' portion of time.
    
    Essentially find the max and min values (of population) in every small interval, and then check whether their difference > min * fluc.

    Inputs:
    - sim: a simulation object
    - interval: int, how many records to take average over,
                and then compare this "local mean" with "whole-tail mean" and expect the difference to be less than fluc.
    - start: (0, 1) float, decides where you expect to check convergence from. Smaller start needs earlier convergence.
    - fluc: (0, 1) float. How much fluctuation is allowed between the average value of a small interval and the mean.
    '''

    if (start < 0) or (start > 1):
        raise ValueError("start should be a float in (0, 1)")
    if (fluc < 0) or (fluc > 1):
        raise ValueError("fluc should be a float in (0, 1)")
    if (type(interval) != int) or (interval < 1):
        raise ValueError("interval should be an int >= 1")
    
    interval = figure_t.scale_interval(interval, sim.compress_itv)

    start_index = int(sim.max_record * start)  # where the tail starts
    num_interval = int((sim.max_record - start_index) / interval)  # how many intervals in total

    # find the max and min value of the small intervals 
    # initiate as average of the first interval
    min_U = np.mean(sim.U[:, :, start_index : start_index + interval])
    max_U = np.mean(sim.U[:, :, start_index : start_index + interval])
    min_V = np.mean(sim.V[:, :, start_index : start_index + interval])
    max_V = np.mean(sim.V[:, :, start_index : start_index + interval])

    for i in range(1, num_interval):
        # lower and upper bound of current interval
        lower = start_index + i * interval
        upper = lower + interval

        ave_U = np.mean(sim.U[:, :, lower : upper])
        ave_V = np.mean(sim.V[:, :, lower : upper])

        # Compare with min, max
        if ave_U > max_U:
            max_U = ave_U
        if ave_U < min_U:
            min_U = ave_U

        if ave_V > max_V:
            max_V = ave_V
        if ave_V < min_V:
            min_V = ave_V

        # check whether (max - min) > min * fluc
        if (max_U - min_U) > min_U * fluc:
            return False
        if (max_V - min_V) > min_V * fluc:
            return False
            
    return True




def combine_sim(sim1, sim2):
    '''
    Combine data of sim1 and sim2. 
    Intended usage: assume sim1 and sim2 has the same N, M, maxtime, interval, boundary, max_record, and I, X, P
    combine_sim then combines the two results and calculate a new weighted average of the two data, return a new sim object. 
    Essentially allows breaking up many rounds of simulations into several smaller pieces, and then put together.

    Inputs:
    - sim1, sim2: both stochastic_model.simulation object. All input parameters the same except for sim_time, print_pct and seed.
            Raises error if not.

    Returns:

    - sim3:       a new simulation object whose U, V, U_pi, V_pi are weighted averages of sim1 and sim2
                (weighted by sim_time). 
                sim3.print_pct is set to sim1's, seed set to None, sim_time set to sum of sim1's and sim2's. All other params same as sim1
    '''
    if not (sim1.N == sim2.N and
            sim1.M == sim2.M and
            sim1.maxtime == sim2.maxtime and
            sim1.record_itv == sim2.record_itv and
            sim1.boundary == sim2.boundary and
            sim1.max_record == sim2.max_record and
            np.array_equal(sim1.I, sim2.I) and
            np.array_equal(sim1.X, sim2.X) and
            np.array_equal(sim1.P, sim2.P)):
        
        raise ValueError('sim1 and sim2 have different input parameters (N, M, maxtime, interval, boundary, max_record, or I, X, P).')

    if sim1.seed == sim2.seed:
        raise ValueError('Cannot combine two simulations with the same seed.')
    
    # copy sim1, except for no data and a different sim_time
    combined_sim_time = sim1.sim_time + sim2.sim_time
    sim3 = sim1.copy(copy_data = False)
    sim3.sim_time = combined_sim_time
    sim3.seed = None

    for i in range(sim3.N):
        for j in range(sim3.M):
            for k in range(sim3.max_record):
                sim3.U[i][j][k] = (sim1.U[i][j][k] * sim1.sim_time + sim2.U[i][j][k] * sim2.sim_time) / combined_sim_time
                sim3.V[i][j][k] = (sim1.V[i][j][k] * sim1.sim_time + sim2.V[i][j][k] * sim2.sim_time) / combined_sim_time
                sim3.U_pi[i][j][k] = (sim1.U_pi[i][j][k] * sim1.sim_time + sim2.U_pi[i][j][k] * sim2.sim_time) / combined_sim_time
                sim3.V_pi[i][j][k] = (sim1.V_pi[i][j][k] * sim1.sim_time + sim2.V_pi[i][j][k] * sim2.sim_time) / combined_sim_time

    return sim3






__version__ = '1.0.0'

'''
version history:

0.1.0: first publishing, May 11, 2025
0.1.1: fix dependency errors
0.1.2: fixing module not find error
0.1.3: restructuring package
0.1.4 ~ 0.1.6: fixing moviepy import issue
0.1.7: changed name back to 'piegy'
0.1.8: updated installation in README
0.1.9: first round of full debugging

1.0.0: first version in PyPI
1.1.0: debugging. Updated a range of functions, in the following modules: figures, videos, test_var, model, figure_tools
1.1.1: minor debugging in model module.

'''


'''
Stores and reads a simulation object.

Functions:
- save_data:    save a simulation object.
- read_data:    read a simulation object.
'''


from . import model as model

import json
import gzip
import os


def save_data(sim, dirs = '', print_msg = True):
    '''
    Saves a simulation object. Data will be stored at dirs/data.json.gz

    Inputs:
    - sim:        Your simulation object.
    - dirs:       Where to save it.
    - print_msg:  Whether to print message after saving.
    '''

    try:
        _ = sim.N
    except AttributeError:
        raise ValueError('sim is not a simulation object')

    if dirs != '':
        # add slash '/'
        if dirs[:-1] != '/':
            dirs += '/'
        if not os.path.exists(dirs):
            os.makedirs(dirs)
        
    data = []
    
    inputs1 = []
    inputs1.append(sim.N)
    inputs1.append(sim.M)
    inputs1.append(sim.maxtime)
    inputs1.append(sim.record_itv)
    inputs1.append(sim.sim_time)
    inputs1.append(sim.boundary)
    inputs1.append(sim.I.tolist())
    inputs1.append(sim.X.tolist())
    inputs1.append(sim.P.tolist())
    data.append(inputs1)

    inputs2 = []
    inputs2.append(sim.print_pct)
    inputs2.append(sim.seed)
    inputs2.append(sim.UV_dtype)
    inputs2.append(sim.pi_dtype)
    data.append(inputs2)

    # skipped rng
    
    outputs = []
    outputs.append(sim.max_record)
    outputs.append(sim.compress_itv)
    outputs.append(sim.U.tolist())
    outputs.append(sim.V.tolist())
    outputs.append(sim.U_pi.tolist())
    outputs.append(sim.V_pi.tolist())
    # H&V_pi_total are not saved, will be calculated when reading the data
    data.append(outputs)
    
    data_json = json.dumps(data)
    data_bytes = data_json.encode('utf-8')
    data_dirs = dirs + 'data.json.gz'
    
    with gzip.open(data_dirs, 'w') as f:
        f.write(data_bytes)
    
    if print_msg:
        print('data saved: ' + data_dirs)



def read_data(dirs):
    '''
    Reads and returns a simulation object.

    Inputs:
    - dirs:       where to read from, just provide the folder-subfolder names. Don't include 'data.json.gz'
    - print_msg:  this function prints a message when the sim.compress_itv != None. Setting print_msg = False will skip ignore this message.

    Returns:
    - sim: a piegy.model.simulation object read from the data.
    '''
    
    if dirs != '':
        # add slash '/'
        if dirs[:-1] != '/':
            dirs += '/'
        if not os.path.exists(dirs):
            raise FileNotFoundError('dirs not found: ' + dirs)
            
    if not os.path.isfile(dirs + 'data.json.gz'):
        raise FileNotFoundError('data not found in ' + dirs)
    
    with gzip.open(dirs + 'data.json.gz', 'r') as f:
        data_bytes = f.read()
    data_json = data_bytes.decode('utf-8')
    data = json.loads(data_json)

    # inputs
    try:
        sim = model.simulation(N = data[0][0], M = data[0][1], maxtime = data[0][2], record_itv = data[0][3],
                            sim_time = data[0][4], boundary = data[0][5], I = data[0][6], X = data[0][7], P = data[0][8], 
                            print_pct = data[1][0], seed = data[1][1], UV_dtype = data[1][2], pi_dtype = data[1][3])
    except:
        raise ValueError('Invalid input parameters saved in data')

    # outputs
    try:
        sim.set_data(data_empty = False, max_record = data[2][0], compress_itv = data[2][1], 
                    U = data[2][2], V = data[2][3], U_pi = data[2][4], V_pi = data[2][5])
    except:
        raise ValueError('Invalid simulation results saved in data')
    
    return sim








'''
Helper functions for making figures.

Functions:
- heatmap:          Make a heatmap based on input data. Sets title, text ... as well
- bar:              Make a barplot. Sets title, text ... as well.
- scatter:          Make a scatter plot. Sets title, text ... as well.
- gen_title:        Generates a title when the plot is about an interval of time.
- gen_text:         Generates a text about standard deviation info.
- scale_interval:   scale interval if sim's data was already reduced.
- ave_interval:     Calculates average value of data over a time interval.
- ave_interval_1D:  Return in a 1D format.
- config_mpl:       Configure Matplotlib parameters in a nice format
'''


import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

# move ax a bit left if add text
# default value is [0.125, 0.11, 0.9, 0.88]


def heatmap(data, ax = None, cmap = "Greens", annot = False, fmt = '.3g', title = None, text = None):
    '''
    Helper function for making heatmaps.

    Inputs:
        data:   1D data for which you want to make a heatmap. 
        ax:     matplotlib ax to plot on. 
        cmap:   Color of heatmap. Uses matplotlib color maps
        annot:  Whether to show numbers of every block.
        fmt:    Number format for annotations. How many digits you want to keep.
        title:  The title you want to add. None means no title.
        text:   Adds some text in a text block at the top-right corner.

    Returns:
        fig:    Seaborn heatmap.
    '''

    if ax == None:
        _, ax = plt.subplots()

    if text != None:
        ax.text(0.66, 0.9, text, size = 10, linespacing = 1.5, transform = plt.gcf().transFigure)

    ax = sns.heatmap(data, ax = ax, cmap = cmap, annot = annot, fmt = fmt)
    ax.set_title(title)
    
    return ax



def bar(data, ax = None, color = "green", xlabel = None, ylabel = None, title = None, text = None):
    '''
    Helper Function for making barplots.

    Inputs:
        data:   2D data to make barplot.
        ax:     matplotlib ax to plot on. 
        color:  Uses Matplotlib colors.
        xlabel, y_label: 
                Label for axes.
        title:  Title for the barplot.
        text:   Adds some text in a text block at the top-right corner.
    
    Returns:
        fig:    A Matplotlib barplot.
    '''

    N = np.array(data).size
    xaxis = np.array([i for i in range(N)])
    
    if ax == None:
        if N > 60:
            # make figure larger if has more data points
            _, ax = plt.subplots(figsize = (min(N * 0.107, 6.4), 4.8))
        else:
            _, ax = plt.subplots()

    if text != None:
        ax.text(0.66, 0.9, text, size = 10, linespacing = 1.5, transform = plt.gcf().transFigure)

    ax.bar(x = xaxis, height = data, color = color)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)
    
    return ax



def scatter(X, Y, ax = None, color = "orange", alpha = 0.25, xlabel = "x", ylabel = "y", title = None):
    '''
    Helper function for makeing scatter plots.

    Inputs:
        X:      x-coordinates of points.
        Y:      y-coordinates of points.
        ax:     matplotlib ax to plot on. 
        Note color is Matplotlib colors.
    
    Returns:
        fig:    A Matplotlib scatter plot.
    '''
    
    if ax == None:
        _, ax = plt.subplots(figsize = (7.2, 5.4))
    ax.scatter(X, Y, color = color, alpha = alpha)
    
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)

    return ax



def gen_title(title, start, end):
    '''
    Generate a title for plot when it's about an interval of time.
    '''
    title += ", " + str(round(start * 100, 1)) + " ~ " + str(round(end * 100, 1)) + "%"
    return title



def gen_text(ave, std):
    '''
    Generate text about standard deviation info.
    '''
    text = "ave=" + str(round(ave, 2)) + ", std=" + str(round(std, 2))
    return text



def ave_interval(data, start_index, end_index):
    '''
    Calculate average value of data over an interval. Return a 2D np.array
    Assume data is 3D with shape N x M x K, then takes average on the 3rd axis.

    Input:
        data:       3D np.array or list. Will take average on the 3rd axis.
        start_index, end_index: 
                    over what interval to take average.

    Returns:
        data_ave:   2D np.array with shape N x M, contains average value of data.
    '''
    
    N = len(data)
    M = len(data[0])
    
    # plot a particular record
    if start_index == end_index:
        start_index = end_index - 1
        
    data_ave = np.zeros((N, M))
    
    for i in range(N):
        for j in range(M):
            for k in range(start_index, end_index):
                data_ave[i][j] += data[i][j][k]
            data_ave[i][j] /= (end_index - start_index)
    
    return data_ave



def ave_interval_1D(data, start_index, end_index):
    '''
    Calculate average value of data over an interval. Return a 1D np.array.
    Assume data is 3D and has shape (1 x M x K) or (N x 1 x K). Then implicitly 'compress' that 1 and takes average on the 3rd axis.

    Input:
        data:       3D np.array or list. One of its dimensions must have size 1. Will take average on the 3rd axis.
        start_index, end_index: 
                    over what interval to take average.

    Returns:
        data_ave:   1D np.array with len N * M, contains average value of data.
    '''
    
    N = len(data)
    M = len(data[0])

    if start_index == end_index:
        start_index = end_index - 1
        
    data_ave = np.zeros(N * M)
    
    for i in range(N):
        for j in range(M):
            for k in range(start_index, end_index):
                data_ave[i * M + j] += data[i][j][k]
            data_ave[i * M + j] /= (end_index - start_index)
    
    return data_ave



def scale_interval(interval, compress_itv):
    # scale interval if sim's data was already reduced.
    if compress_itv < 1:
        raise ValueError('figures.scale_interval has compress_itv < 1:', compress_itv)

    interval = int(interval / compress_itv)
    if interval == 0:
        print('Warning: data already smoothed by an interval: sim.compress_itv =', compress_itv, 'which is coarser than your', interval)
        interval = 1

    return interval



def config_mpl(mpl):
    '''
    Configure Matplotlib figures
    '''
    mpl.rcParams['savefig.dpi'] = 300
    
    mpl.rcParams['font.family'] = 'serif'
    mpl.rcParams['font.serif'] = plt.rcParams['font.serif']
    mpl.rcParams['lines.linewidth'] = 1.75
    mpl.rcParams['font.size'] = 11
    mpl.rcParams['axes.labelsize'] = 12
    mpl.rcParams['xtick.major.size'] = 10
    mpl.rcParams['ytick.major.size'] = 9
    mpl.rcParams['xtick.minor.size'] = 4
    mpl.rcParams['ytick.minor.size'] = 4

    mpl.rcParams['xtick.major.width'] = 1.5
    mpl.rcParams['ytick.major.width'] = 1.5
    mpl.rcParams['xtick.minor.width'] = 1.5
    mpl.rcParams['ytick.minor.width'] = 1.5





'''
Main Module of Stochastic Model
-------------------------------

Contains all the necessary tools to build a model and run simulations based on Gillespie Algorithm.

Classes:
- patch:        (Private) Simulates a single patch in the N x M space. Assume no spatial structure within a patch.
                All spatial movements are based on patches.
- simulation:   Stores input parameters and data generated during simulation.


Functions:
- find_nb_zero_flux:        (Private) Return pointers to a patch's neighbors under zero-flux (no-boundary) boundary condition.
- find_nb_periodical:       (Private) Return pointers to a patch's neighbors under periodical (with-booundary) boundary condition.
- find_event:               (Private) Pick a random event to happen.
- make_signal_zero_flux:    (Private) Expand event index (return value of find_event) to a detailed signal, under zero-flux boundary condition.
- make_signal_periodical:   (Private) Expand event index (return value of find_event) to a detailed signal, under periodical boundary condition.
- single_init:              (Private) Initialize a single simulation. Meaning of 'single': see single_test and run.
- single_test:              (Private) Run a single simulation.
                            'single' means a single round of simulation. You can run many single rounds and then take the average --- that's done by <run> function.
- run:                      Run multiple simulations and then take the average. All the simulation will use the same parameters.
                            Set a seed for reproducible results.
- demo_model:               Returns a demo model (a simulation object).

NOTE: Only simulation class and run function are intended for direct usages.
'''


import math
import numpy as np
from timeit import default_timer as timer


# data type used by simulation.U and simulation.V
UV_DTYPE = 'float32'

# data type used by simulation.U_pi and V_pi
PI_DTYPE = 'float64'

# data type for storing rates in single_test an single_init
RATES_DTYPE = 'float64'


class patch:
    '''
    A single patch in the N x M space.
    Interacts with neighboring patches, assuming no spatial structure within a patch.
    Initialized in single_init function.

    Class Functions:

        __init__:
            Inputs:
                U, V:       initial value of U and V
                matrix:     payoff matrix for U and V. The canonical form is 2x2, here we ask for a flattened 1x4 form.
                patch_var:  np.array of [mu1, mu2, w1, w2, kappa1, kappa2]
        
        __str__:
            Print patch object in a nice way.

        set_nb_pointers:
            Set pointers to neighbors of this patch object.

        update_pi:
            Update U_pi, V_pi and payoff rates (payoff rates are the first two numbers in self.pi_death_rates).
        
        update_k:
            Update natural death rates (the last two numbers in self.pi_death_rates).

        update_mig:
            Update migration rates.

        get_pi_death_rates, get_mig_rates:
            Return respective members.

        change_popu:
            Change U, V based on input signal.
    '''
    
    def __init__(self, U, V, matrix = [-0.1, 0.4, 0, 0.2], patch_var = [0.5, 0.5, 100, 100, 0.001, 0.001]):

        self.U = U                          # int, U population. Initialized upon creating object.
        self.V = V                          # int, V population
        self.U_pi = 0                       # float, payoff
        self.V_pi = 0

        self.matrix = matrix                # np.array or list, len = 4, payoff matrix
        self.mu1 = patch_var[0]             # float, how much proportion of the population migrates (U) each time
        self.mu2 = patch_var[1]
        self.w1 = patch_var[2]              # float, strength of payoff-driven effect. Larger w <=> stronger payoff-driven motion
        self.w2 = patch_var[3]
        self.kappa1 = patch_var[4]          # float, carrying capacity, determines death rates
        self.kappa2 = patch_var[5]

        self.nb = None                                  # list of patch objects (pointers), point to neighbors, initialized seperatedly (after all patches are created)
        self.pi_death_rates = [0 for _ in range(4)]     # list, len = 4, rates of payoff & death
                                                        # first two are payoff rates, second two are death rates
        self.mig_rates = [0 for _ in range(8)]          # list, len = 8, migration rates, stored in an order: up, down, left, right, 
                                                        # first 4 are U's mig_rate, the last 4 are V's
        self.sum_pi_death_rates = 0                     # float, sum of pi_death_rates
        self.sum_mig_rates = 0                          # float, sum of mig_rates
        

    def __str__(self):
        self_str = ''
        self_str += 'U, V = ' + str(self.U) + ', ' + str(self.V) + '\n'
        self_str += 'pi = ' + str(self.U_pi) + ', ' + str(self.V_pi) + '\n'
        self_str += 'matrix = ' + str(self.matrix) + '\n'
        self_str += 'mu1, mu2 = ' + str(self.mu1) + ', ' + str(self.mu2) + '\n'
        self_str += 'w1, w2 = ' + str(self.w1) + ', ' + str(self.w2) + '\n'
        self_str += 'kappa1, kappa2 = ' + str(self.kappa1) + ', ' + str(self.kappa2) + '\n'
        self_str += '\n'
        self_str += 'nb = ' + str(self.nb)
        self_str += 'pi_death_rates = ' + str(self.pi_death_rates) + '\n'
        self_str += 'mig_rates = ' + str(self.mig_rates) + '\n'
        self_str += 'sum_pi_death_rates = ' + str(self.sum_pi_death_rates) + '\n'
        self_str += 'sum_mig_rates = ' + str(self.sum_mig_rates) + '\n'

        return self_str
    
    
    def set_nb_pointers(self, nb):
        # nb is a list of pointers (point to patches)
        # nb is passed from the simulation class
        self.nb = nb
    
    
    def update_pi_k(self):
        # calculate payoff and natural death rates

        U = self.U  # bring the values to front
        V = self.V
        sum_minus_1 = U + V - 1  # this value is used several times
        
        if sum_minus_1 > 0:
            # interaction happens only if there is more than 1 individual
            
            if U != 0:
                # no payoff if U == 0
                self.U_pi = (U - 1) / sum_minus_1 * self.matrix[0] + V / sum_minus_1 * self.matrix[1]
            else:
                self.U_pi = 0
                
            if V != 0:
                self.V_pi = U / sum_minus_1 * self.matrix[2] + (V - 1) / sum_minus_1 * self.matrix[3]
            else:
                self.V_pi = 0
                
        else:
            # no interaction, hence no payoff, if only 1 individual
            self.U_pi = 0
            self.V_pi = 0

        # update payoff rates
        self.pi_death_rates[0] = abs(U * self.U_pi)
        self.pi_death_rates[1] = abs(V * self.V_pi)

        # update natural death rates
        self.pi_death_rates[2] = self.kappa1 * U * (sum_minus_1 + 1)
        self.pi_death_rates[3] = self.kappa2 * V * (sum_minus_1 + 1)

        # update sum of rates
        self.sum_pi_death_rates = sum(self.pi_death_rates)
    

    def update_mig(self):
        # calculate migration rates

        # store the 'weight' of migration, i.e. value of f/g functions for neighbors
        U_weight = [0, 0, 0, 0]
        V_weight = [0, 0, 0, 0]

        for i in range(4):
            if self.nb[i] != None:
                U_weight[i] = 1 + pow(math.e, self.w1 * self.nb[i].U_pi)
                V_weight[i] = 1 + pow(math.e, self.w2 * self.nb[i].V_pi)

        mu1_U = self.mu1 * self.U
        mu2_V = self.mu2 * self.V

        mu1_U_divide_sum = mu1_U / sum(U_weight)
        mu2_V_divide_sum = mu2_V / sum(V_weight)
        
        for i in range(4):
            self.mig_rates[i] = mu1_U_divide_sum * U_weight[i]
            self.mig_rates[i + 4] = mu2_V_divide_sum * V_weight[i]

        # update sum of rates
        self.sum_mig_rates = mu1_U + mu2_V
    

    def get_sum_rates(self):
        # return sum of all 12 rates
        return self.sum_pi_death_rates + self.sum_mig_rates


    def find_event(self, expected_sum):
        # find the event within the 12 events based on expected sum-of-rates within this patch

        if expected_sum < self.sum_pi_death_rates:
            # in the first 4 events (payoff and death events)
            event = 0
            current_sum = 0
            while current_sum < expected_sum:
                current_sum += self.pi_death_rates[event]
                event += 1
            event -= 1

        else:
            # in the last 8 events (migration events):
            event = 0
            current_sum = self.sum_pi_death_rates
            while current_sum < expected_sum:
                current_sum += self.mig_rates[event]
                event += 1
            event += 3  # i.e., -= 1, then += 4 (to account for the first 4 payoff & death rates)

        return event

    
    def change_popu(self, s):
        # convert s (a signal, passed from simulation class) to a change in population
        
        # s = 0, 1, 2 are for U
        # s = 0 for migration IN, receive an immigrant
        if s == 0:
            self.U += 1       # receive an immigrant
        # s = 1 for migration OUT / death due to carrying capacity
        elif s == 1:
            if self.U > 0:
                self.U -= 1
        # s = 2 for natural birth / death, due to payoff
        elif s == 2:
            if self.U_pi > 0:
                self.U += 1   # natural growth due to payoff
            elif self.U > 0:
                self.U -= 1   # natural death due to payoff
        
        # s = 3, 4, 5 are for V
        elif s == 3:
            self.V += 1
        elif s == 4:
            if self.V > 0:
                self.V -= 1
        else:
            if self.V_pi > 0:
                self.V += 1
            elif self.V > 0:
                self.V -= 1
                


class simulation:
    '''
    Store simulation data and input parameters.
    Initialize a simulation object to run simulations.

    Public Class Functions:

        __init__:
            Create a simulation object. Also initialize data storage.

        __str__:
            Print simulation object in a nice way.

        copy:
            Return a deep copy of self. Can choose whether to copy data as well. Default is to copy.

        clear:
            clear all data stored, set U, V, U_pi, V_pi to zero arrays

        change_maxtime:
            Changes maxtime of self. Update data storage as well.

        set_seed:
            Set a new seed. 

        compress:
            compress data by only storing average values
    '''

    def __init__(self, N, M, maxtime, record_itv, sim_time, boundary, I, X, P, print_pct = 25, seed = None, UV_dtype = UV_DTYPE, pi_dtype = PI_DTYPE):

        self.check_valid_input(N, M, maxtime, record_itv, sim_time, boundary, I, X, P, print_pct, seed)
        
        self.N = N                      # int, N x M is spatial dimension
        self.M = M                      # int, can't be 1. If want to make 1D space, use N = 1. And this model doesn't work for 1x1 space (causes NaN)
        self.maxtime = maxtime          # float or int, run simulation for how long time
        self.record_itv = record_itv        # float, record data every record_itv of time
        self.sim_time = sim_time        # int, run this many of rounds (of single_test)
        self.boundary = boundary        # bool, the N x M space have boundary or not (i.e., zero-flux (True) or periodical (False))
        self.I = np.array(I)            # N x M x 2 np.array, initial population. Two init-popu for every patch (U and V)
        self.X = np.array(X)            # N x M x 4 np.array, matrices. The '4' comes from 2x2 matrix flattened to 1D
        self.P = np.array(P)            # N x M x 6 np.array, 'patch variables', i.e., mu1&2, w1&2, kappa1&2
        self.print_pct = print_pct      # int, print how much percent is done, need to be non-zero
        self.seed = seed                # non-negative int, seed for random generator
        self.UV_dtype = UV_dtype        # what data type to store population, should be a float format. This value is passed to np.array. 
                                        # Default is 'float32', use lower accuracy to reduce data size.
        self.pi_dtype = pi_dtype        # what data type to store payoff, should be a float format. This value is passed to np.array. 
                                        # Default is 'float64'

        self.init_storage()                         # initialize storage bins. Put in a separate function because might want to change maxtime 
                                                    # and that doesn't need to initialze the whole object again


    def init_storage(self):
        # initialize storage bins
        self.data_empty = True                      # whether data storage bins are empty. model.run will refuse to run (raise error) if not empty.
        self.max_record = int(self.maxtime / self.record_itv)                         # int, how many data points to store sin total
        self.compress_itv = 1                       # int, intended to reduce size of data (if not 1). Updated by compress_data function
                                                    # if set to an int, say 20, sim will take average over every 20 data points and save them as new data.
                                                    # May be used over and over again to recursively reduce data size. 
                                                    # Default is 1, not to take average.
        self.U = np.zeros((self.N, self.M, self.max_record), dtype = self.UV_dtype)      # N x M x max_record np.array, float32, stores population of U in every patch over tiem
        self.V = np.zeros((self.N, self.M, self.max_record), dtype = self.UV_dtype)
        self.U_pi = np.zeros((self.N, self.M, self.max_record), dtype = self.pi_dtype)   # similar to U, but for U's payoff and float 64
        self.V_pi = np.zeros((self.N, self.M, self.max_record), dtype = self.pi_dtype)


    def check_valid_input(self, N, M, maxtime, record_itv, sim_time, boundary, I, X, P, print_pct, seed):
        # check whether the inputs are valid
        # seed, UV_dtype, pi_dtype is handled by numpy

        if (N < 1) or (M < 1):
            raise ValueError('N < 1 or M < 1')
        if (N == 1) and (M == 1):
            raise ValueError('Model fails for 1x1 space')
        if (M == 1):
            raise ValueError('Please set N = 1 for 1D space.')
        if maxtime <= 0:
            raise ValueError('Please set a positive number for maxtime')
        if record_itv <= 0:
            raise ValueError('Please set a positive number for record_itv')
        if sim_time <= 0:
            raise ValueError('Please set a positive number for sim_time')
        if type(boundary) != bool:
            raise TypeError('boundary not a bool. Please use True for zero-flux (with boundary) or False for periodical (no boundary)')
        
        if (type(I) != list) and (type(I) != np.ndarray):
            raise TypeError('Please set I as a list or np.ndarray')
        if np.array(I).shape != (N, M, 2):
            raise ValueError('Please set I as a N x M x 2 shape list or array. 2 is for init values of U, V at every patch')
        
        if (type(X) != list) and (type(X) != np.ndarray):
            raise TypeError('Please set X as a list or np.ndarray')
        if np.array(X).shape != (N, M, 4):
            raise ValueError('Please set X as a N x M x 4 shape list or array. 4 is for the flattened 2x2 payoff matrix')
        
        if (type(P) != list) and (type(P) != np.ndarray):
            raise TypeError('Please set P as a list or np.ndarray')
        if np.array(P).shape != (N, M, 6):
            raise ValueError('Please set P as a N x M x 6 shape list or array. 6 is for mu1, mu2, w1, w2, kappa1, kappa2')
        
        if print_pct <= 0:
            raise ValueError('Please use an int > 0 for print_pct or None for not printing progress.')
        
        # seed is checked by number
        

    def check_valid_data(self, data_empty, max_record, compress_itv, U, V, U_pi, V_pi):
        # check whether a set of data is valid, used when reading a saved model
        if type(data_empty) != bool:
            raise TypeError('data_empty not a bool')
        
        if type(max_record) != int:
            raise TypeError('max_record not an int')
        if max_record < 0:
            raise ValueError('max_record < 0')
        
        if type(compress_itv) != int:
            raise TypeError('compress_itv not an int')
        if compress_itv < 0:
            raise ValueError('compress_itv < 0')
   

    def __str__(self):
        # print this sim in a nice format

        self_str = ''
        self_str += 'N = ' + str(self.N) + '\n'
        self_str += 'M = ' + str(self.M) + '\n'
        self_str += 'maxtime = ' + str(self.maxtime) + '\n'
        self_str += 'record_itv = ' + str(self.record_itv) + '\n'
        self_str += 'sim_time = ' + str(self.sim_time) + '\n'
        self_str += 'boundary = ' + str(self.boundary) + '\n'
        self_str += 'print_pct = ' + str(self.print_pct) + '\n'
        self_str += 'seed = ' + str(self.seed) + '\n'
        self_str += 'UV_dtype = \'' + self.UV_dtype + '\'\n'
        self_str += 'pi_dtype = \'' + self.pi_dtype + '\'\n'
        self_str += 'compress_itv = ' + str(self.compress_itv) + '\n'
        self_str += '\n'

        # check whether I, X, P all same (compare all patches to (0, 0))
        I_same = True
        X_same = True
        P_same = True
        for i in range(self.N):
            for j in range(self.M):
                for k in range(2):
                    if self.I[i][j][k] != self.I[0][0][k]:
                        I_same = False
                for k in range(4):
                    if self.X[i][j][k] != self.X[0][0][k]:
                        X_same = False
                for k in range(6):
                    if self.P[i][j][k] != self.P[0][0][k]:
                        P_same = False
        
        if I_same:
            self_str += 'I all same: ' + str(self.I[0][0]) + '\n'
        else:
            self_str += 'I:\n'
            for i in range(self.N):
                for j in range(self.M):
                    self_str += str(self.I[i][j]) + ' '
                self_str += '\n'
            self_str += '\n'
        
        if X_same:
            self_str += 'X all same: ' + str(self.X[0][0]) + '\n'
        else:
            self_str += 'X:\n'
            for i in range(self.N):
                for j in range(self.M):
                    self_str += str(self.X[i][j]) + ' '
                self_str += '\n'
            self_str += '\n'

        if P_same:
            self_str += 'P all same: ' + str(self.P[0][0]) + '\n'
        else:
            self_str += 'P:\n'
            for i in range(self.N):
                for j in range(self.M):
                    self_str += str(self.P[i][j]) + ' '
                self_str += '\n'

        return self_str
    

    def copy(self, copy_data = True):
        # return deep copy of self
        # copy_data decides whether to copy data as well
        if type(copy_data) != bool:
            raise TypeError('Please give a bool as argument: whether to copy data or not')

        sim2 = simulation(N = self.N, M = self.M, maxtime = self.maxtime, record_itv = self.record_itv, sim_time = self.sim_time, boundary = self.boundary,
                          I = np.copy(self.I), X = np.copy(self.X), P = np.copy(self.P), 
                          print_pct = self.print_pct, seed = self.seed, UV_dtype = self.UV_dtype, pi_dtype = self.pi_dtype)

        if copy_data:
            # copy data as well
            sim2.set_data(self.data_empty, self.max_record, self.compress_itv, self.U, self.V, self.U_pi, self.V_pi)

        return sim2


    def calculate_ave(self):
        # get the average value over sim_time many simulations
        if self.sim_time != 1:
            for i in range(self.N):
                for j in range(self.M):
                    for t in range(self.max_record):
                        self.U[i][j][t] /= self.sim_time
                        self.V[i][j][t] /= self.sim_time
                        self.U_pi[i][j][t] /= self.sim_time
                        self.V_pi[i][j][t] /= self.sim_time


    def change_maxtime(self, maxtime):
        # change maxtime
        if (type(maxtime) != float) and (type(maxtime) != int):
            raise TypeError('Please pass in a float or int as the new maxtime.')
        if maxtime <= 0:
            raise ValueError('Please use a positive maxtime.')
        self.maxtime = maxtime
        self.init_storage()


    def set_seed(self, seed):
        # set seed
        self.seed = seed

    
    def reset_data(self):
        # clear all data stored, set all to 0
        self.init_storage()


    def clear(self):
        # clear data by simply reseting
        self.reset_data()


    def set_data(self, data_empty, max_record, compress_itv, U, V, U_pi, V_pi):
        # set data to the given data values
        # copies are made
        self.check_valid_data(data_empty, max_record, compress_itv, U, V, U_pi, V_pi)

        self.data_empty = data_empty
        self.max_record = max_record
        self.compress_itv = compress_itv
        self.U = np.copy(U)
        self.V = np.copy(V)
        self.U_pi = np.copy(U_pi)
        self.V_pi = np.copy(V_pi)


    def compress(self, compress_itv = 5):
        # compress data by only storing average values
        
        if type(compress_itv) != int:
            raise TypeError('Please use an int as compress_itv')
        if compress_itv < 1:
            raise ValueError('Please use record_itv >= 1')
        if compress_itv == 1:
            return
        
        self.compress_itv *= compress_itv  # may be reduced over and over again
        self.max_record = int(self.max_record / compress_itv)  # number of data points after reducing

        U_reduced = np.zeros((self.N, self.M, self.max_record), dtype = self.UV_dtype)
        V_reduced = np.zeros((self.N, self.M, self.max_record), dtype = self.UV_dtype)
        U_pi_reduced = np.zeros((self.N, self.M, self.max_record), dtype = self.pi_dtype)
        V_pi_reduced = np.zeros((self.N, self.M, self.max_record), dtype = self.pi_dtype)

        for i in range(self.N):
            for j in range(self.M):
                for k in range(self.max_record):
                    lower = k * compress_itv  # upper and lower bound of current record_itv
                    upper = lower + compress_itv
                    U_reduced[i][j][k] = np.mean(self.U[i, j, lower : upper])
                    V_reduced[i][j][k] = np.mean(self.V[i, j, lower : upper])
                    U_pi_reduced[i][j][k] = np.mean(self.U_pi[i, j, lower : upper])
                    V_pi_reduced[i][j][k] = np.mean(self.V_pi[i, j, lower : upper])
        
        self.U = U_reduced
        self.V = V_reduced
        self.U_pi = U_pi_reduced
        self.V_pi = V_pi_reduced




def find_nb_zero_flux(N, M, i, j):
    '''
    Find neighbors of patch (i, j) in zero-flux boundary condition. i.e., the space is square with boundary.
    Return neighbors' indices in an order: up, down, left, right.
    Index will be None if no neighbor exists in that direction.
    '''
    nb_indices = []
    
    if i != 0: 
        nb_indices.append([i - 1, j])  # up
    else:
        nb_indices.append(None)    # neighbor doesn't exist
    
    if i != N - 1:
        nb_indices.append([i + 1, j])  # down
    else:
        nb_indices.append(None)
    
    if j != 0:
        nb_indices.append([i, j - 1])  # left
    else:
        nb_indices.append(None)
    
    if j != M - 1:
        nb_indices.append([i, j + 1])  # right
    else:
        nb_indices.append(None)
    
    return nb_indices




def find_nb_periodical(N, M, i, j):
    '''
    Find neighbors of patch (i, j) in periodical boundary condition. i.e., the space is a sphere.
    Return neighbors' indices in an order: up, down, left, right.
    If space not 1D, a neighbor always exists.
    If space is 1D, say N = 1, we don't allow (0, j) to migrate up & down (self-self migration is considered invalid)
    '''
    nb_indices = []
    
    # up
    if N != 1:
        if i != 0: 
            nb_indices.append([i - 1, j])
        else:
            nb_indices.append([N - 1, j])
    else:
        nb_indices.append(None)  # can't migrate to itself
    
    # down
    if N != 1:
        if i != N - 1:
            nb_indices.append([i + 1, j])
        else:
            nb_indices.append([0, j])
    else:
        nb_indices.append(None)
    
    # left
    # No need to check M == 1 because we explicitly asked for M > 1
    if j != 0:
        nb_indices.append([i, j - 1])
    else:
        nb_indices.append([i, M - 1])
    
    # right
    if j != M - 1:
        nb_indices.append([i, j + 1])
    else:
        nb_indices.append([i, 0])
        
    return nb_indices




def find_patch(expected_sum, patch_rates, sum_rates_by_row, sum_rates):
    '''
    Find which patch the event is in. Only patch index is found, patch.find_event find which event it is exactly.

    Inputs:
        expected_sum:       a random number * sum of all rates. Essentially points to a random event.
                            We want to find the patch that contains this pointer.
        patch_rates:        a N x M np.array. Stores sum of the 12 rates in every patch.
        sum_rates_by_row:   a 1D np.array with len = N. Stores the sum of the M x 12 rates in every row.
        sum_rates:          sum of all N x M x 12 rates.
    
    Returns:
        row, col:           row and column number of where the patch.
    '''
    
    # Find row first
    if expected_sum < sum_rates / 2:
        # search row forwards if in the first half of rows
        current_sum = 0
        row = 0
        while current_sum < expected_sum:
            current_sum += sum_rates_by_row[row]
            row += 1
        row -= 1
        current_sum -= sum_rates_by_row[row]  # need to subtract that row (which caused current sum to exceed expected_sum)
    else:
        # search row backwards if in the second half of rows
        current_sum = sum_rates
        row = len(patch_rates) - 1
        while current_sum > expected_sum:
            current_sum -= sum_rates_by_row[row]
            row -= 1
        row += 1
        # don't need subtraction here, as current_sum is already < expected same

    # Find col in that row
    if (expected_sum - current_sum) < sum_rates_by_row[row] / 2:
        # search col forwards if in the first half of that row
        col = 0
        while current_sum < expected_sum:
            current_sum += patch_rates[row][col]
            col += 1
        col -= 1
        current_sum -= patch_rates[row][col]  # need a subtraction
    else:
        # search col backwards if in the second half of that row
        current_sum += sum_rates_by_row[row]
        col = len(patch_rates[0]) - 1
        while current_sum > expected_sum:
            current_sum -= patch_rates[row][col]
            col -= 1
        col += 1
        # don't need subtraction

    return row, col, current_sum




def make_signal_zero_flux(i, j, e):
    '''
    Find which patch to change what based on i, j, e (event number) value, for the zero-flux boundary condition

    Inputs: 
        i, j is the position of the 'center' patch, e is which event to happen there.
        Another patch might be influenced as well if a migration event was picked.

        Possible values for e:
            e = 0 or 1: natural change of U/V due to payoff. 
                        Can be either brith or death (based on payoff is positive or negative).
                        Cooresponds to s = 2 or 5 in the patch class
            e = 2 or 3: death of U/V due to carrying capacity. 
                        Cooresponds to s = 1 or 4 in patch: make U/V -= 1
            e = 4 ~ 7:  migration events of U, patch (i, j) loses an individual, and another patch receives one.
                        we use the up-down-left-right rule for the direction. 4 means up, 5 means down, ...
                        Cooresponds to s = 0 for the mig-in patch (force U += 1), and s = 1 for the mig-out patch (force U -= 1)
            e = 8 ~ 11: migration events of V.
                        Cooresponds to s = 3 for the mig-in patch (force V += 1), and s = 4 for the mig-out patch (force V -= 1)
    '''
    if e < 6:
        if e == 0:
            return [[i, j, 2]]
        elif e == 1:
            return [[i, j, 5]]
        elif e == 2:
            return [[i, j, 1]]
        elif e == 3:
            return [[i, j, 4]]
        elif e == 4:
            return [[i, j, 1], [i - 1, j, 0]]
        else:
            return [[i, j, 1], [i + 1, j, 0]]
    else:
        if e == 6:
            return [[i, j, 1], [i, j - 1, 0]]
        elif e == 7:
            return [[i, j, 1], [i, j + 1, 0]]
        elif e == 8:
            return [[i, j, 4], [i - 1, j, 3]]
        elif e == 9:
            return [[i, j, 4], [i + 1, j, 3]]
        elif e == 10:
            return [[i, j, 4], [i, j - 1, 3]]
        elif e == 11:
            return [[i, j, 4], [i, j + 1, 3]]
        else:
            raise RuntimeError('A bug in code: invalid event number encountered:', e)  # debug line



def make_signal_periodical(N, M, i, j, e):
    '''
    Find which patch to change what based on i, j, e value, for the periodical boundary condition
    Similar to make_signal_zero_flux.
    '''
    
    if e < 6:
        if e == 0:
            return [[i, j, 2]]
        elif e == 1:
            return [[i, j, 5]]
        elif e == 2:
            return [[i, j, 1]]
        elif e == 3:
            return [[i, j, 4]]
        elif e == 4:
            if i != 0:
                return [[i, j, 1], [i - 1, j, 0]]
            else:
                return [[i, j, 1], [N - 1, j, 0]]
        else:
            if i != N - 1:
                return [[i, j, 1], [i + 1, j, 0]]
            else:
                return [[i, j, 1], [0, j, 0]]
    else:
        if e == 6:
            if j != 0:
                return [[i, j, 1], [i, j - 1, 0]]
            else:
                return [[i, j, 1], [i, M - 1, 0]]
        elif e == 7:
            if j != M - 1:
                return [[i, j, 1], [i, j + 1, 0]]
            else:
                return [[i, j, 1], [i, 0, 0]]
        elif e == 8:
            if i != 0:
                return [[i, j, 4], [i - 1, j, 3]]
            else:
                return [[i, j, 4], [N - 1, j, 3]]
        elif e == 9:
            if i != N - 1:
                return [[i, j, 4], [i + 1, j, 3]]
            else:
                return [[i, j, 4], [0, j, 3]]
        elif e == 10:
            if j != 0:
                return [[i, j, 4], [i, j - 1, 3]]
            else:
                return [[i, j, 4], [i, M - 1, 3]]
        elif e == 11:
            if j != M - 1:
                return [[i, j, 4], [i, j + 1, 3]]
            else:
                return [[i, j, 4], [i, 0, 3]]
        else:
            raise RuntimeError('A bug in code: invalid event number encountered:', e)  # debug line




def nb_need_change(ni, signal):
    '''
    Check whether a neighbor needs to change.
    Two cases don't need change: either ni is None (doesn't exist) or in signal (is a last-change patch and already updated)

    Inputs:
        ni:         index of a neighbor, might be None if patch doesn't exist.
        signal:     return value of make_signal_zero_flux or make_signal_periodical.
    
    Returns:
        True or False, whether the neighboring patch specified by ni needs change
    '''
    
    if ni == None:
        return False
    
    for si in signal:
        if ni[0] == si[0] and ni[1] == si[1]:
            return False
    
    return True




def single_init(sim, rng):
    '''
    The first major function for the model.
    Initialize all variables and run 1 round, then pass variables and results to single_test.

    Input:
        sim is a simulation object
        rng is random number generator (np.random.default_rng), initialized by model.run
    '''

    #### Initialize Data Storage ####

    world = [[patch(sim.I[i][j][0], sim.I[i][j][1], sim.X[i][j], sim.P[i][j]) for j in range(sim.M)] for i in range(sim.N)]  # N x M patches
    patch_rates = np.zeros((sim.N, sim.M), dtype = RATES_DTYPE)  # every patch's sum-of-12-srates
    sum_rates_by_row = np.zeros((sim.N), dtype = RATES_DTYPE)  # every row's sum-of-patch, i.e., sum of 12 * M rates in every row.
    sum_rates = 0  # sum of all N x M x 12 rates

    signal = None

    nb_indices = None
    if sim.boundary:
        nb_indices = [[find_nb_zero_flux(sim.N, sim.M, i, j) for j in range(sim.M)] for i in range(sim.N)]
    else:
        nb_indices = [[find_nb_periodical(sim.N, sim.M, i, j) for j in range(sim.M)] for i in range(sim.N)]
        
    for i in range(sim.N):
        for j in range(sim.M):
            nb = []
            for k in range(4):
                if nb_indices[i][j][k] != None:
                    # append a pointer to the patch
                    nb.append(world[nb_indices[i][j][k][0]][nb_indices[i][j][k][1]])
                else:
                    # nb doesn't exist
                    nb.append(None)
            # pass it to patch class and store
            world[i][j].set_nb_pointers(nb)
    

    #### Begin Running ####

    # initialize payoff & natural death rates
    for i in range(sim.N):
        for j in range(sim.M):
            world[i][j].update_pi_k()
    
    # initialize migration rates & the rates list
    for i in range(sim.N):
        for j in range(sim.M):
            world[i][j].update_mig()
            # store rates & sum of rates
            patch_rates[i][j] = world[i][j].get_sum_rates()
        sum_rates_by_row[i] = sum(patch_rates[i])

    sum_rates = sum(sum_rates_by_row)
    
    # pick the first random event
    expected_sum = rng.random() * sum_rates
    # find patch first
    i0, j0, current_sum = find_patch(expected_sum, patch_rates, sum_rates_by_row, sum_rates)
    # then find which event in that patch
    e0 = world[i0][j0].find_event(expected_sum - current_sum)

    # initialize signal
    if sim.boundary:
        signal = make_signal_zero_flux(i0, j0, e0)   # walls around world
    else:
        signal = make_signal_periodical(sim.N, sim.M, i0, j0, e0)  # no walls around world
    
    # change U&V based on signal
    for si in signal:
        world[si[0]][si[1]].change_popu(si[2])
        
    # time increment
    time = (1 / sum_rates) * math.log(1 / rng.random())
    
    # record
    if time > sim.record_itv:
        record_index = int(time / sim.record_itv)
        for i in range(sim.N):
            for j in range(sim.M):
                for k in range(record_index):
                    sim.U[i][j][k] += world[i][j].U
                    sim.V[i][j][k] += world[i][j].V
                    sim.U_pi[i][j][k] += world[i][j].U_pi
                    sim.V_pi[i][j][k] += world[i][j].V_pi
                    # we simply add to that entry, and later divide by sim_time to get the average (division in run function)
    
    return time, world, nb_indices, patch_rates, sum_rates_by_row, sum_rates, signal




def single_test(sim, front_info, end_info, update_sum_frequency, rng):
    '''
    Runs a single simulation, from time = 0 to sim.maxtime.
    run recursively calls single_test to get the average data. 

    Inputs:
        sim:                    a simulation object, created by user and carries all parameters & storage bins.
        front_info, end_info:   passed by run to show messages, like the current round number in run. Not intended for direct usages.
        update_sum_frequency:   re-calculate sums this many times in simulation. 
                                Our sums are gradually updated over time. So might have precision errors for large maxtime.
        rng:                    np.random.default_rng. Initialized by model.run
    '''

    # initialize helper variables
    # used to print progress, i.e., how much percent is done
    one_time = sim.maxtime / max(100, update_sum_frequency) 
    one_progress = 0
    if sim.print_pct != None:
        # print progress, x%
        print(front_info +  ' 0%' + end_info, end = '\r')
        one_progress = sim.maxtime * sim.print_pct / 100
    else:
        one_progress = 2 * sim.maxtime     # not printing

    # our sums (sum_rates_by_row and sum_rates) are gradually updated over time. This may have precision errors for large maxtime.
    # So re-sum everything every some percentage of maxtime.
    one_update_sum = sim.maxtime / update_sum_frequency

    current_time = one_time
    current_progress = one_progress
    current_update_sum = one_update_sum

    max_record = int(sim.maxtime / sim.record_itv)

    
    # initialize
    time, world, nb_indices, patch_rates, sum_rates_by_row, sum_rates, signal = single_init(sim, rng)
    record_index = int(time / sim.record_itv)
    # record_time is how much time has passed since the last record
    # if record_time > record_itv:
    #    we count how many record_itvs are there in record_time, denote the number by multi_records
    #    then store the current data in multi_records number of cells in the list
    #    and subtract record_time by the multiple of record_itv, so that record_time < record_itv
    record_time = time - record_index * sim.record_itv

    ### Large while loop ###
    
    while time < sim.maxtime:
    
        # print progress & correct error of sum_rates
        if time > current_time:
            # a new 1% of time
            current_time += one_time
            if time > current_progress:
                # print progress
                print(front_info +  ' ' + str(round(time / sim.maxtime * 100)) + '%' + end_info, end = '\r')
                current_progress += one_progress

            if time > current_update_sum:
                current_update_sum += one_update_sum
                for i in range(sim.N):
                    sum_rates_by_row[i] = sum(patch_rates[i])
                sum_rates = sum(sum_rates_by_row)

        
        # before updating last-changed patches, subtract old sum of rates (so as to update sum of rates by adding new rates later)
        for si in signal:
            # si[0] is row number, si[1] is col number
            old_patch_rate = world[si[0]][si[1]].get_sum_rates()
            sum_rates_by_row[si[0]] -= old_patch_rate
            sum_rates -= old_patch_rate

        # update last-changed patches
        # update payoff and death rates first
        for si in signal:
            world[si[0]][si[1]].update_pi_k()
        # then update migration rates, as mig_rates depend on neighbor's payoff
        for si in signal:
            world[si[0]][si[1]].update_mig()
            
            # update rates stored
            new_patch_rate = world[si[0]][si[1]].get_sum_rates()
            # update patch_rates
            patch_rates[si[0]][si[1]] = new_patch_rate
            # update sum_rate_by_row and sum_rates_by_row by adding new rates
            sum_rates_by_row[si[0]] += new_patch_rate
            sum_rates += new_patch_rate

        # update neighbors of last-changed patches
        for si in signal:
            for ni in nb_indices[si[0]][si[1]]:
                # don't need to update if the patch is a last-change patch itself or None
                # use helper function to check
                if nb_need_change(ni, signal):
                    # update migratino rates
                    world[ni[0]][ni[1]].update_mig()
                    # Note: no need to update patch_rates and sum of rates, as update_mig doesn't change total rates in a patch.
                    # sum_mig_rate is decided by mu1 * U + mu2 * V, and pi_death_rate is not changed.
               
        # pick the first random event
        expected_sum = rng.random() * sum_rates
        # find patch first
        i0, j0, current_sum = find_patch(expected_sum, patch_rates, sum_rates_by_row, sum_rates)
        # then find which event in that patch
        e0 = world[i0][j0].find_event(expected_sum - current_sum)

        # make signal
        if sim.boundary:
            signal = make_signal_zero_flux(i0, j0, e0)
        else:
            signal = make_signal_periodical(sim.N, sim.M, i0, j0, e0)

        # let the event happen
        for si in signal:
            world[si[0]][si[1]].change_popu(si[2])
        
        # increase time
        r1 = rng.random()
        dt = (1 / sum_rates) * math.log(1 / r1)
        time += dt
        record_time += dt
    
        if time < sim.maxtime:
            # if not exceeds maxtime
            if record_time > sim.record_itv:
                multi_records = int(record_time / sim.record_itv)
                record_time -= multi_records * sim.record_itv

                for i in range(sim.N):
                    for j in range(sim.M):
                        for k in range(record_index, record_index + multi_records):
                            sim.U[i][j][k] += world[i][j].U
                            sim.V[i][j][k] += world[i][j].V
                            sim.U_pi[i][j][k] += world[i][j].U_pi
                            sim.V_pi[i][j][k] += world[i][j].V_pi
                record_index += multi_records
        else:
            # if already exceeds maxtime
            for i in range(sim.N):
                for j in range(sim.M):
                    for k in range(record_index, max_record):
                        sim.U[i][j][k] += world[i][j].U
                        sim.V[i][j][k] += world[i][j].V
                        sim.U_pi[i][j][k] += world[i][j].U_pi
                        sim.V_pi[i][j][k] += world[i][j].V_pi

    ### Large while loop ends ###
    
    if sim.print_pct != None:
        print(front_info + ' 100%' + ' ' * 20, end = '\r')  # empty spaces to overwrite predicted runtime




def run(sim, predict_runtime = False, message = ''):
    '''
    Main function. Recursively calls single_test to run many simulations and then takes the average.

    Inputs:
    - sim is a simulation object.
    - predict_runtime = False will not predict how much time still needed, set to True if you want to see.
    - message is used by some functions in figures.py to print messages.
    '''

    if not sim.data_empty:
        raise RuntimeError('sim has non-empty data')
    
    start = timer()   # runtime

    sim.data_empty = False
    rng = np.random.default_rng(sim.seed)
    
    # passed to single_test to print progress
    if sim.print_pct == 0:
        sim.print_pct = 5    # default print_pct

    update_sum_frequency = 4   # re-calculate sums this many times. See input desciption of single_test
    
    ### simulations ###
    i = 0
    
    while i < sim.sim_time:
        # use while loop so that can go backwards if got numerical issues

        end_info = ''
        if predict_runtime:
            if i > 0:
                time_elapsed = timer() - start
                pred_runtime = time_elapsed / i * (sim.sim_time - i)
                end_info = ', ~' + str(round(pred_runtime, 2)) + 's left'
        
        front_info = ''
        if sim.print_pct != None:
            front_info = message + 'round ' + str(i) + ':'
            print(front_info + ' ' * 30, end = '\r')  # the blank spaces are to overwrite percentages, e.g. 36 %
        
        try:
            single_test(sim, front_info, end_info, update_sum_frequency, rng)
            i += 1
        except IndexError:
            update_sum_frequency *= 4
            print('Numerical issue at round ' + str(i) + '. Trying higher precision now. See doc if err repeats')
            # not increasing i: redo current round.
        
    ### simulations end ###

    sim.calculate_ave()
    
    stop = timer()
    print(' ' * 30, end = '\r')  # overwrite all previous prints
    print(message + 'runtime: ' + str(round(stop - start, 2)) + ' s')
    return




def demo_model():
    '''
    Returns a demo model.simulation object
    '''

    N = 10                  # Number of rows
    M = 10                  # Number of cols
    maxtime = 300           # how long you want the model to run
    record_itv = 0.1        # how often to record data.
    sim_time = 1            # repeat simulation to reduce randomness
    boundary = True         # boundary condition.

    # initial population for the N x M patches. 
    I = [[[44, 22] for _ in range(M)] for _ in range(N)]
    
    # flattened payoff matrices, total resource is 0.4, cost of fighting is 0.1
    X = [[[-0.1, 0.4, 0, 0.2] for _ in range(M)] for _ in range(N)]
    
    # patch variables
    P = [[[0.5, 0.5, 200, 200, 0.001, 0.001] for _ in range(M)] for _ in range(N)]

    print_pct = 5           # print progress
    seed = 36               # seed for random number generation
    UV_dtype = 'float32'    # data type for population
    pi_dyna = 'float64'    # data type for payoff

    # create a simulation object
    sim = simulation(N, M, maxtime, record_itv, sim_time, boundary, I, X, P, 
                            print_pct = print_pct, seed = seed, UV_dtype = UV_dtype, pi_dtype = pi_dyna)
    
    return sim





'''
Contains all the major plot functions. 

Plots for population:
- UV_heatmap:       Used for 2D space (both N, M > 1), plot distribution of U, V in all patches within a specified time interval.
                    Average population over that interval is taken.
- UV_bar:           Used for 1D space (N or M == 1), counterpart of UV_heatmap.
                    Plot average distribution of U, V in a specified time interval in a barplot.
- UV_dyna:         Plot change of total U, V overtime.
- UV_hist:          Make a histogram of U, V in a specified time interval.
- UV_std:           Plot change of standard deviation of U, V over time.
- UV_expected_val:  Calculate expected U, V distribution based on matrices, assuming no migration return values (np.array)
- UV_expected:      Calculate expected distribution of U, V based on matrices, assuming no migration.


Plots for payoff:
- pi_heatmap:       Used for 2D space, plot distribution of U_pi & V_pi within a specified time interval.
                    Average payoff over that interval is taken.
- pi_bar:           Used for 1D space, counterpart of pi_heatmap.
                    Plot average distribution of U_pi & V_pi in a specified time interval in a bar plot.
- pi_dyna:         Plot change of total U_pi, V_pi overtime.
- pi_hist:          Make a histogram of U_pi, V_pi in a specified time interval.
- pi_std:           Plot change of standard deviation of U_pi, V_pi over time.


Popu-payoff correlation:
- UV_pi:            Make two scatter plots: x-axes are U, V, y-axes are U's and V's payoff, in a specified time interval.
                    Reveals relationship between population and payoff.

'''


from .tools import figure_tools as figure_t
from . import model as model

import matplotlib.pyplot as plt
import numpy as np


# curve type in plot
# used by UV_dyna, UV_std, and pi_dyna
CURVE_TYPE = '-'



def UV_heatmap(sim, ax_U = None, ax_V = None, U_color = 'Purples', V_color = 'Greens', start = 0.95, end = 1.0, annot = False, fmt = '.3g'):
    '''
    Makes two heatmaps for U, V average distribution over a time interval, respectively. Works best for 2D space.
    1D works as well, but figures look bad.

    Inputs:
        sim:        A model.simulation object.
        ax_U, ax_V: matplotlib axes to plot on. New axes will be created if None is given.
        U_color:    Color for U's heatmap, uses matplotlib color maps.
        V_color:    Color for V's heatmap.
        start:      (0,1) float, where the interval should start from. Intended as a 'percentage'. 
                    For example, start = 0.8 means the interval should start from the 80% point of sim.maxtime.
        end:        (0,1) float, where the interval ends.
        annot:      bool, whether to add annotations (show exact numbers for every patch).
        fmt:        Number format for annotations. How many digits you want to keep. Please set annot = True first and then use fmt.

    Returns:
        ax_U, ax_V: matplotlib axes with heatmaps of U, V distribution plotted upon.
    '''
    
    start_index = int(start * sim.max_record)
    end_index = int(end * sim.max_record)
    
    # see ave_interval below
    U_ave = figure_t.ave_interval(sim.U, start_index, end_index)
    V_ave = figure_t.ave_interval(sim.V, start_index, end_index)
    
    #### plot ####
    
    U_title = figure_t.gen_title('U', start, end)
    U_text = figure_t.gen_text(np.mean(U_ave), np.std(U_ave))
    V_title = figure_t.gen_title('V', start, end)
    V_text = figure_t.gen_text(np.mean(V_ave), np.std(V_ave))

    ax_U = figure_t.heatmap(U_ave, ax_U, U_color, annot, fmt, U_title, U_text)
    ax_V = figure_t.heatmap(V_ave, ax_V, V_color, annot, fmt, V_title, V_text)
        
    return ax_U, ax_V
    


def UV_bar(sim, ax_U = None, ax_V = None, U_color = 'purple', V_color = 'green', start = 0.95, end = 1.0):
    '''
    Makes two barplots for U, V average distribution over a time interval. Works best for 1D space.
    2D works as well, but figures look bad.

    Inputs:
        sim:        A model.simulation object.
        ax_U, ax_V: matplotlib axes to plot on. New axes will be created if None is given.
        U_color:    Color of U's barplot. Uses Matplotlib colors.
                    See available colors at: https://matplotlib.org/stable/gallery/color/named_colors.html
        V_color:    Color of V's barplot. Uses Matplotlib colors.
        start:      (0,1) float. How much proportion of sim.maxtime you want the interval to start from.
        end:        (0,1) float. Where you want the interval to end.

    Returns:
        ax_U, ax_V: matplotlib axes with bar plots for U and V plotted upon.
    '''
    
    start_index = int(start * sim.max_record)
    end_index = int(end * sim.max_record)
    
    U_ave = figure_t.ave_interval_1D(sim.U, start_index, end_index)
    V_ave = figure_t.ave_interval_1D(sim.V, start_index, end_index)

    #### plot ####

    U_title = figure_t.gen_title('U', start, end)
    U_text = figure_t.gen_text(np.mean(U_ave), np.std(U_ave))
    V_title = figure_t.gen_title('V', start, end)
    V_text = figure_t.gen_text(np.mean(V_ave), np.std(V_ave))

    ax_U = figure_t.bar(U_ave, ax = ax_U, color = U_color, xlabel = 'patches', ylabel = 'U', title = U_title, text = U_text)
    ax_V = figure_t.bar(V_ave, ax = ax_V, color = V_color, xlabel = 'patches', ylabel = 'V', title = V_title, text = V_text)

    return ax_U, ax_V




def UV_dyna(sim, ax = None, interval = 20, grid = True):
    '''
    Plots how total U, V change overtime.
    The curves are not directly based on every single data point. 
    Rather, it takes the average over many intervals of points to smooth out local fluctuations.
        For example, interval = 20 means the first point on the curves are based on the average value of data points 0~19.
        So if there are 2000 data points in total, then there will be 2000 / 20 = 100 points on the curves.

    Inputs:
        sim:        A model.simulation object.
        ax:         matplotlib ax to plot on. New ax will be created if None is given.
        interval:   How many data points to take average over. Larger value makes curves smoother, but also loses local fluctuations.
                    NOTE: this interval doesn't overlap with sim.compress_itv. 
                    e.g. you already took average over every 20 data points, then using interval <= 20 here has no smoothing effect.
        grid:       Whether to add grid lines to plot.
    
    Returns:
        ax:        matplotlib ax, contains U's, V's, and sum of U & V population.
    '''
    
    # store the average values in lists
    U_curve = []
    V_curve = []
    total_curve = []

    interval = figure_t.scale_interval(interval, sim.compress_itv)
    interval_num = int(sim.max_record / interval)
    
    for i in range(interval_num):
        U_ave = figure_t.ave_interval(sim.U, i * interval, (i + 1) * interval)
        V_ave = figure_t.ave_interval(sim.V, i * interval, (i + 1) * interval)
        
        U_curve.append(np.sum(U_ave))
        V_curve.append(np.sum(V_ave))
        total_curve.append(U_curve[-1] + V_curve[-1])
        
    #### plot ####   
    xaxis = np.linspace(0, sim.maxtime, len(U_curve))

    if ax == None:
        _, ax = plt.subplots()
    ax.grid(grid)
    ax.plot(xaxis, U_curve, CURVE_TYPE, label = 'U')
    ax.plot(xaxis, V_curve, CURVE_TYPE, label = 'V')
    ax.plot(xaxis, total_curve, CURVE_TYPE, label = 'total')
    ax.set_xlabel('Time')
    ax.set_ylabel('Population')
    ax.set_title('Population U & V Over Time')
    ax.legend()

    return ax




def UV_hist(sim, ax_U = None, ax_V = None, U_color = 'purple', V_color = 'green', start = 0.95, end = 1.0):
    '''
    Makes density histograms for U, V's average distribution over an interval.
    Sometimes it may not be shown in density plots due to matplotlib features.

    Returns:
        ax_U, ax_V: matplotlib axes with heatmaps of U, V population density plotted upon.
    '''

    start_index = int(start * sim.max_record)
    end_index = int(end * sim.max_record)
    
    U_ave = figure_t.ave_interval_1D(sim.U, start_index, end_index)
    V_ave = figure_t.ave_interval_1D(sim.V, start_index, end_index)
    
    #### plot ####
    
    if ax_U == None:
        _, ax_U = plt.subplots()
    ax_U.set_xlabel('Population U')
    ax_U.set_ylabel('Density')
    ax_U.hist(U_ave, color = U_color, density = True)
    ax_U.set_title(figure_t.gen_title('U Hist', start, end))
    
    if ax_V == None:
        _, ax_V = plt.subplots()
    ax_V.set_xlabel('Population V')
    ax_V.set_ylabel('Density')
    ax_V.hist(V_ave, color = V_color, density = True)
    ax_V.set_title(figure_t.gen_title('V Hist', start, end))

    return ax_U, ax_V




def UV_std(sim, ax = None, interval = 20, grid = True):
    '''
    Plots how standard deviation of U, V change over time.
    Takes average over many small interval to smooth out local fluctuations.

    Returns:
        ax:    matplotlib ax, contains U's and V's std curves.
    '''

    interval = figure_t.scale_interval(interval, sim.compress_itv)
    interval_num = int(sim.max_record / interval)
    
    U_std = []
    V_std = []
    
    for i in range(interval_num):
        U_ave = figure_t.ave_interval(sim.U, i * interval, (i + 1) * interval)
        V_ave = figure_t.ave_interval(sim.V, i * interval, (i + 1) * interval)
        
        U_std.append(np.std(U_ave))
        V_std.append(np.std(V_ave))
    
    #### plot ####
    xaxis = np.linspace(0, sim.maxtime, len(U_std))

    if ax == None:
        _, ax = plt.subplots()
    ax.grid(grid)
    ax.plot(xaxis, U_std, CURVE_TYPE, label = 'U std')
    ax.plot(xaxis, V_std, CURVE_TYPE, label = 'V std')
    ax.legend()
    ax.set_xlabel('Time')
    ax.set_ylabel('Std Dev')
    ax.set_title('Population Std-Dev Dynamics')

    return ax



def UV_expected_val(sim):
    '''
    Calculate expected U & V distribution based on matrices, assume no migration.
    To differentiate from UV_expected in figures.py: this one return arrays (values).
    '''
    
    U_expected = np.zeros((sim.N, sim.M))
    V_expected = np.zeros((sim.N, sim.M))
    
    for i in range(sim.N):
        for j in range(sim.M):
            # say matrix = [a, b, c, d]
            # U_proportion = (d - b) / (a - b - c + d)
            U_prop = (sim.X[i][j][3] - sim.X[i][j][1]) / (sim.X[i][j][0] - sim.X[i][j][1] - sim.X[i][j][2] + sim.X[i][j][3])
            # equilibrium payoff, U_payoff = V_payoff
            eq_payoff = U_prop * sim.X[i][j][0] + (1 - U_prop) * sim.X[i][j][1]
            
            # payoff / kappa * proportion
            U_expected[i][j] = eq_payoff / sim.P[i][j][4] * U_prop
            V_expected[i][j] = eq_payoff / sim.P[i][j][5] * (1 - U_prop)
                
    return U_expected, V_expected




def UV_expected(sim, ax_U = None, ax_V = None, U_color = 'Purples', V_color = 'Greens', annot = False, fmt = '.3g'):
    '''
    Calculate expected population distribution based on matrices, assuming no migration.
    For the formulas, see stochastic_mode.expected_UV

    Some Inputs:
        Note the colors are color maps.
    
    Returns:
    ax_U, ax_V: If 2D (N and M both > 1), then ax_U and ax_V are heatmaps.
                If 1D (N or M == 1), then ax_U and ax_V are barplots.
    '''
    
    U_expected, V_expected = UV_expected_val(sim)
    
    U_text = figure_t.gen_text(np.mean(U_expected), np.std(U_expected))
    V_text = figure_t.gen_text(np.mean(V_expected), np.std(V_expected))
    
    #### plot ####
    
    if (sim.N != 1) and (sim.M != 1):
        # 2D
        ax_U = figure_t.heatmap(U_expected, ax_U, U_color, annot, fmt, title = 'Expected U', text = U_text)
        ax_V = figure_t.heatmap(V_expected, ax_V, V_color, annot, fmt, title = 'Expected V', text = V_text)

    else:
        # 1D     
        ax_U = figure_t.bar(U_expected.flatten(), ax_U, color = U_color, xlabel = 'patches', ylabel = 'popu', title = 'Expected Population U', text = U_text)
        ax_V = figure_t.bar(V_expected.flatten(), ax_V, color = V_color, xlabel = 'patches', ylabel = 'popu', title = 'Expected Population V', text = V_text)

    return ax_U, ax_V




def pi_heatmap(sim, ax_U = None, ax_V = None, U_color = 'BuPu', V_color = 'YlGn', start = 0.95, end = 1.0, annot = False, fmt = '.3g'):
    '''
    Make heatmaps for payoff in a specified interval.
    Works best for 2D. 1D works as well, but figures look bad.

    Some Inputs:.
        Note the colors are matplotlib color maps.

    Returns:
        ax_U, ax_V: Seaborn heatmaps, for U's & V's payoff distribution, respectively.
    '''
    
    start_index = int(sim.max_record * start)
    end_index = int(sim.max_record * end)
    
    U_pi_ave = figure_t.ave_interval(sim.U_pi, start_index, end_index)
    V_pi_ave = figure_t.ave_interval(sim.V_pi, start_index, end_index)
    
    U_title = figure_t.gen_title('Payoff ' + r'$p_U$', start, end)
    U_text = figure_t.gen_text(np.mean(U_pi_ave), np.std(U_pi_ave))
    V_title = figure_t.gen_title('Payoff ' + r'$p_V$', start, end)
    V_text = figure_t.gen_text(np.mean(V_pi_ave), np.std(V_pi_ave))
    
    ax_U = figure_t.heatmap(U_pi_ave, ax_U, U_color, annot, fmt, U_title, U_text)
    ax_V = figure_t.heatmap(V_pi_ave, ax_V, V_color, annot, fmt, V_title, V_text)

    return ax_U, ax_V




def pi_bar(sim, ax_U = None, ax_V = None, U_color = 'violet', V_color = 'yellowgreen', start = 0.95, end = 1.0):
    '''
    Make barplot for payoff in a specified interval.
    Works best for 1D. 2D works as well, but figures look bad.

    Returns:
        ax_U, ax_V: matplotlib axes with barplots of U and V payoff distribution plotted upon.
    '''
    
    start_index = int(sim.max_record * start)
    end_index = int(sim.max_record * end)
    
    U_pi_ave = figure_t.ave_interval_1D(sim.U_pi, start_index, end_index)
    V_pi_ave = figure_t.ave_interval_1D(sim.V_pi, start_index, end_index)
    
    U_title = figure_t.gen_title(r'$p_U$', start, end)
    U_text = figure_t.gen_text(np.mean(U_pi_ave), np.std(U_pi_ave))
    V_title = figure_t.gen_title(r'$p_V$', start, end)
    V_text = figure_t.gen_text(np.mean(V_pi_ave), np.std(V_pi_ave))
    
    ax_U = figure_t.bar(U_pi_ave, ax_U, U_color, 'Patches', 'Payoff ' + r'$p_U$', U_title, U_text)
    ax_V = figure_t.bar(V_pi_ave, ax_V, V_color, 'Patches', 'Payoff ' + r'$p_V$', V_title, V_text)

    return ax_U, ax_V




def pi_dyna(sim, ax = None, interval = 20, grid = True):
    '''
    Plot how payoffs change over time.

    Returns:
        ax:    matplotlib ax of U's, V's, and sum of U & V payoff.
    '''
    
    U_curve = []
    V_curve = []
    total_curve = []

    interval = figure_t.scale_interval(interval, sim.compress_itv)
    interval_num = int(sim.max_record / interval)
    
    for i in range(interval_num):
        U_ave = figure_t.ave_interval(sim.U_pi, i * interval, (i + 1) * interval)
        V_ave = figure_t.ave_interval(sim.V_pi, i * interval, (i + 1) * interval)
    
        U_curve.append(np.sum(U_ave))
        V_curve.append(np.sum(V_ave))
        total_curve.append(U_curve[-1] + V_curve[-1])
        
    #### plot ####    
    xaxis = np.linspace(0, sim.maxtime, len(U_curve))
    
    if ax == None:
        _, ax = plt.subplots()
    ax.grid(grid)
    ax.plot(xaxis, U_curve, CURVE_TYPE, label = r'$p_U$')
    ax.plot(xaxis, V_curve, CURVE_TYPE, label = r'$p_V$')
    ax.plot(xaxis, total_curve, CURVE_TYPE, label = 'total')
    ax.set_xlim(0, sim.maxtime)
    ax.set_xlabel('Time')
    ax.set_ylabel('Payoff')
    ax.set_title('Payoff ' + r'$p_U$' + ' & ' + r'$p_V$' + ' over time')
    ax.legend()

    return ax




def pi_hist(sim, ax_U = None, ax_V = None, U_color = 'violet', V_color = 'yellowgreen', start = 0.95, end = 1.0):
    '''
    Makes deensity histograms of U's and V's payoffs in a sepcified interval.
    Sometimes it may not be shown in density plots due to matplotlib features.
    
    Returns:
        ax_U, ax_V:     histogram of U's and V's payoff.
    '''

    start_index = int(start * sim.max_record)
    end_index = int(end * sim.max_record)

    U_pi_ave = figure_t.ave_interval_1D(sim.U_pi, start_index, end_index)
    V_pi_ave = figure_t.ave_interval_1D(sim.V_pi, start_index, end_index)
    
    #### plot ####
    
    if ax_U == None:
        _, ax_U = plt.subplots()
    ax_U.set_xlabel('Payoff ' + r'$p_U$')
    ax_U.set_ylabel('Density')
    ax_U.hist(U_pi_ave, color = U_color, density = True)
    ax_U.set_title(figure_t.gen_title('Payoff ' + r'$p_U$' + ' Hist', start, end))
    
    if ax_V == None:
        _, ax_V = plt.subplots()
    ax_V.set_xlabel('Payoff ' + r'$p_V$')
    ax_V.set_ylabel('Density')
    ax_V.hist(V_pi_ave, color = V_color, density = True)
    ax_V.set_title(figure_t.gen_title('Payoff ' + r'$p_V$' + ' Hist', start, end))

    return ax_U, ax_V




def pi_std(sim, ax = None, interval = 20, grid = True):
    '''
    Plots how standard deviation of payoff change over time.

    Returns:
        ax:    matplotlib ax of the std of payoffs.
    '''
    
    
    interval = figure_t.scale_interval(interval, sim.compress_itv)
    interval_num = int(sim.max_record / interval)
    
    U_pi_std = []
    V_pi_std = []
    
    for i in range(interval_num):
        U_pi_ave = figure_t.ave_interval(sim.U_pi, i * interval, (i + 1) * interval)
        V_pi_ave = figure_t.ave_interval(sim.V_pi, i * interval, (i + 1) * interval)
        
        U_pi_std.append(np.std(U_pi_ave))
        V_pi_std.append(np.std(V_pi_ave))
    
    #### plot ####
    xaxis = np.linspace(0, sim.maxtime, len(U_pi_std))
    
    if ax == None:
        _, ax = plt.subplots()
    ax.grid(grid)
    ax.plot(xaxis, U_pi_std, CURVE_TYPE, label = r'$p_U$' + ' std')
    ax.plot(xaxis, V_pi_std, CURVE_TYPE, label = r'$p_V$' + ' std')
    ax.legend()
    ax.set_xlabel('Time')
    ax.set_ylabel('Std Dev')
    ax.set_title('Payoff Std-Dev Dynamics')
    
    return ax




def UV_pi(sim, ax_U = None, ax_V = None, U_color = 'violet', V_color = 'yellowgreen', alpha = 0.4, start = 0.95, end = 1.0):
    '''
    Make two scatter plots: x-axes are population and y-axes are payoff in a specified time interval.
    Reveals relationship between population and payoff.

    Returns:
        ax_U, ax_V: matplotlib axes with U and V population-payoff scatter plots.
    '''
    
    start_index = int(start * sim.max_record)
    end_index = int(end * sim.max_record)
    
    U_ave = figure_t.ave_interval_1D(sim.U, start_index, end_index)
    V_ave = figure_t.ave_interval_1D(sim.V, start_index, end_index)

    U_pi_ave = figure_t.ave_interval(sim.U_pi, start_index, end_index)
    V_pi_ave = figure_t.ave_interval(sim.V_pi, start_index, end_index)
    
    
    ax_U = figure_t.scatter(U_ave, U_pi_ave, ax_U, U_color, alpha, xlabel = 'U', ylabel = 'Payoff ' + r'$p_U$', title = 'U - ' + r'$p_U$')
    ax_V = figure_t.scatter(V_ave, V_pi_ave, ax_V, V_color, alpha, xlabel = 'V', ylabel = 'Payoff ' + r'$p_V$', title = 'V - ' + r'$p_V$')
    
    return ax_U, ax_V


